<link rel="stylesheet" href="./sly/css/app/widgets/sidebar/style.css" />

<sly-sidebar :options="data.{{{widget.widget_id}}}.options" :sidebar-collapsed.sync="state.{{{widget.widget_id}}}.sidebarCollapsed">
  <div slot="sidebar">{{{widget._left_content}}}</div>
  <div>{{{widget._right_content}}}</div>
</sly-sidebar>
