.tailwind *, .tailwind ::before, .tailwind ::after {
  --tw-border-spacing-x: 0;
  --tw-border-spacing-y: 0;
  --tw-translate-x: 0;
  --tw-translate-y: 0;
  --tw-rotate: 0;
  --tw-skew-x: 0;
  --tw-skew-y: 0;
  --tw-scale-x: 1;
  --tw-scale-y: 1;
  --tw-pan-x:  ;
  --tw-pan-y:  ;
  --tw-pinch-zoom:  ;
  --tw-scroll-snap-strictness: proximity;
  --tw-gradient-from-position:  ;
  --tw-gradient-via-position:  ;
  --tw-gradient-to-position:  ;
  --tw-ordinal:  ;
  --tw-slashed-zero:  ;
  --tw-numeric-figure:  ;
  --tw-numeric-spacing:  ;
  --tw-numeric-fraction:  ;
  --tw-ring-inset:  ;
  --tw-ring-offset-width: 0px;
  --tw-ring-offset-color: #fff;
  --tw-ring-color: rgb(59 130 246 / 0.5);
  --tw-ring-offset-shadow: 0 0 #0000;
  --tw-ring-shadow: 0 0 #0000;
  --tw-shadow: 0 0 #0000;
  --tw-shadow-colored: 0 0 #0000;
  --tw-blur:  ;
  --tw-brightness:  ;
  --tw-contrast:  ;
  --tw-grayscale:  ;
  --tw-hue-rotate:  ;
  --tw-invert:  ;
  --tw-saturate:  ;
  --tw-sepia:  ;
  --tw-drop-shadow:  ;
  --tw-backdrop-blur:  ;
  --tw-backdrop-brightness:  ;
  --tw-backdrop-contrast:  ;
  --tw-backdrop-grayscale:  ;
  --tw-backdrop-hue-rotate:  ;
  --tw-backdrop-invert:  ;
  --tw-backdrop-opacity:  ;
  --tw-backdrop-saturate:  ;
  --tw-backdrop-sepia:
}

.tailwind ::backdrop{
  --tw-border-spacing-x: 0;
  --tw-border-spacing-y: 0;
  --tw-translate-x: 0;
  --tw-translate-y: 0;
  --tw-rotate: 0;
  --tw-skew-x: 0;
  --tw-skew-y: 0;
  --tw-scale-x: 1;
  --tw-scale-y: 1;
  --tw-pan-x:  ;
  --tw-pan-y:  ;
  --tw-pinch-zoom:  ;
  --tw-scroll-snap-strictness: proximity;
  --tw-gradient-from-position:  ;
  --tw-gradient-via-position:  ;
  --tw-gradient-to-position:  ;
  --tw-ordinal:  ;
  --tw-slashed-zero:  ;
  --tw-numeric-figure:  ;
  --tw-numeric-spacing:  ;
  --tw-numeric-fraction:  ;
  --tw-ring-inset:  ;
  --tw-ring-offset-width: 0px;
  --tw-ring-offset-color: #fff;
  --tw-ring-color: rgb(59 130 246 / 0.5);
  --tw-ring-offset-shadow: 0 0 #0000;
  --tw-ring-shadow: 0 0 #0000;
  --tw-shadow: 0 0 #0000;
  --tw-shadow-colored: 0 0 #0000;
  --tw-blur:  ;
  --tw-brightness:  ;
  --tw-contrast:  ;
  --tw-grayscale:  ;
  --tw-hue-rotate:  ;
  --tw-invert:  ;
  --tw-saturate:  ;
  --tw-sepia:  ;
  --tw-drop-shadow:  ;
  --tw-backdrop-blur:  ;
  --tw-backdrop-brightness:  ;
  --tw-backdrop-contrast:  ;
  --tw-backdrop-grayscale:  ;
  --tw-backdrop-hue-rotate:  ;
  --tw-backdrop-invert:  ;
  --tw-backdrop-opacity:  ;
  --tw-backdrop-saturate:  ;
  --tw-backdrop-sepia:
}

.tailwind *,
.tailwind ::before,
.tailwind ::after {
  box-sizing: border-box;
  border-width: 0;
  border-style: solid;
  border-color: #e5e7eb;
}

 .tailwind hr {
	 height: 0;
	 color: inherit;
	 border-top-width: 1px;
}
 .tailwind a {
	 color: inherit;
	 text-decoration: inherit;
}
 .tailwind b, .tailwind strong {
	 font-weight: bolder;
}
 .tailwind small {
	 font-size: 80%;
}
 .tailwind table {
	 text-indent: 0;
	 border-color: inherit;
	 border-collapse: collapse;
}
 .tailwind button, .tailwind input, .tailwind optgroup, .tailwind select, .tailwind textarea {
	 font-family: inherit;
	 font-feature-settings: inherit;
	 font-variation-settings: inherit;
	 font-size: 100%;
	 font-weight: inherit;
	 line-height: inherit;
	 color: inherit;
	 margin: 0;
	 padding: 0;
}
 .tailwind button, .tailwind select {
	 text-transform: none;
}
 .tailwind button, .tailwind [type='button'], .tailwind [type='reset'], .tailwind [type='submit'] {
	 -webkit-appearance: button;
	 background-color: transparent;
	 background-image: none;
}
 .tailwind :-moz-focusring {
	 outline: auto;
}
 .tailwind :-moz-ui-invalid {
	 box-shadow: none;
}
 .tailwind progress {
	 vertical-align: baseline;
}
 .tailwind ::-webkit-inner-spin-button, .tailwind ::-webkit-outer-spin-button {
	 height: auto;
}
 .tailwind [type='search'] {
	 -webkit-appearance: textfield;
	 outline-offset: -2px;
}
 .tailwind ::-webkit-search-decoration {
	 -webkit-appearance: none;
}
 .tailwind ::-webkit-file-upload-button {
	 -webkit-appearance: button;
	 font: inherit;
}
 .tailwind summary {
	 display: list-item;
}
 .tailwind blockquote, .tailwind dl, .tailwind dd, .tailwind h1, .tailwind h2, .tailwind h3, .tailwind h4, .tailwind h5, .tailwind h6, .tailwind hr, .tailwind figure, .tailwind p, .tailwind pre {
	 margin: 0;
}
 .tailwind fieldset {
	 margin: 0;
	 padding: 0;
}
 .tailwind legend {
	 padding: 0;
}
 .tailwind ol, .tailwind ul, .tailwind menu {
	 list-style: none;
	 margin: 0;
	 padding: 0;
}
 .tailwind dialog {
	 padding: 0;
}
 .tailwind textarea {
	 resize: vertical;
}
 .tailwind input::placeholder, .tailwind textarea::placeholder {
	 opacity: 1;
	 color: #9ca3af;
}
 .tailwind button, .tailwind [role="button"] {
	 cursor: pointer;
}
 .tailwind :disabled {
	 cursor: default;
}
 .tailwind img, .tailwind svg, .tailwind video, .tailwind canvas, .tailwind audio, .tailwind iframe, .tailwind embed, .tailwind object {
	 display: block;
	 vertical-align: middle;
}
 .tailwind img, .tailwind video {
	 max-width: 100%;
	 height: auto;
}
 .tailwind [hidden] {
	 display: none;
}
 .tailwind .fixed {
	 position: fixed;
}
 .tailwind .absolute {
	 position: absolute;
}
 .tailwind .relative {
	 position: relative;
}
 .tailwind .sticky {
	 position: sticky;
}
 .tailwind .bottom-0 {
	 bottom: 0px;
}
 .tailwind .left-0 {
	 left: 0px;
}
 .tailwind .left-2 {
	 left: 0.5rem;
}
 .tailwind .left-2\.5 {
	 left: 0.625rem;
}
 .tailwind .right-0 {
	 right: 0px;
}
 .tailwind .right-2 {
	 right: 0.5rem;
}
 .tailwind .right-3 {
	 right: 0.75rem;
}
 .tailwind .top-0 {
	 top: 0px;
}
 .tailwind .top-1\/2 {
	 top: 50%;
}
 .tailwind .top-2 {
	 top: 0.5rem;
}
 .tailwind .top-2\.5 {
	 top: 0.625rem;
}
 .tailwind .z-10 {
	 z-index: 10;
}
 .tailwind .z-\[60\] {
	 z-index: 60;
}
 .tailwind .\!mb-0 {
	 margin-bottom: 0px !important;
}
 .tailwind .\!mt-0 {
	 margin-top: 0px !important;
}
 .tailwind .-ml-2 {
	 margin-left: -0.5rem;
}
 .tailwind .-mt-0 {
	 margin-top: 0px;
}
 .tailwind .-mt-0\.5 {
	 margin-top: -0.125rem;
}
 .tailwind .-mt-1 {
	 margin-top: -0.25rem;
}
 .tailwind .mb-1 {
	 margin-bottom: 0.25rem;
}
 .tailwind .mb-5 {
	 margin-bottom: 1.25rem;
}
 .tailwind .ml-0 {
	 margin-left: 0px;
}
 .tailwind .ml-0\.5 {
	 margin-left: 0.125rem;
}
 .tailwind .ml-1 {
	 margin-left: 0.25rem;
}
 .tailwind .ml-2 {
	 margin-left: 0.5rem;
}
 .tailwind .mr-1 {
	 margin-right: 0.25rem;
}
 .tailwind .mr-1\.5 {
	 margin-right: 0.375rem;
}
 .tailwind .mt-0 {
	 margin-top: 0px;
}
 .tailwind .mt-0\.5 {
	 margin-top: 0.125rem;
}
 .tailwind .mt-5 {
	 margin-top: 1.25rem;
}
 .tailwind .box-content {
	 box-sizing: content-box;
}
 .tailwind .\!flex {
	 display: flex !important;
}
 .tailwind .flex {
	 display: flex;
}
 .tailwind .table {
	 display: table;
}
 .tailwind .hidden {
	 display: none;
}
 .tailwind .h-3 {
	 height: 0.75rem;
}
 .tailwind .h-4 {
	 height: 1rem;
}
 .tailwind .h-\[2px\] {
	 height: 2px;
}
 .tailwind .max-h-\[80vh\] {
	 max-height: 80vh;
}
 .tailwind .w-3 {
	 width: 0.75rem;
}
 .tailwind .w-56 {
	 width: 14rem;
}
 .tailwind .w-\[16px\] {
	 width: 16px;
}
 .tailwind .w-\[50px\] {
	 width: 50px;
}
 .tailwind .w-full {
	 width: 100%;
}
 .tailwind .flex-1 {
	 flex: 1 1 0%;
}
 .tailwind .flex-none {
	 flex: none;
}
 .tailwind .-translate-y-1\/2 {
	 --tw-translate-y: -50%;
	 transform: translate(var(--tw-translate-x), var(--tw-translate-y)) rotate(var(--tw-rotate)) skewX(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y));
}
 .tailwind .cursor-default {
	 cursor: default;
}
 .tailwind .cursor-pointer {
	 cursor: pointer;
}
 .tailwind .flex-col {
	 flex-direction: column;
}
 .tailwind .items-center {
	 align-items: center;
}
 .tailwind .justify-between {
	 justify-content: space-between;
}
 .tailwind .gap-0 {
	 gap: 0px;
}
 .tailwind .gap-2 {
	 gap: 0.5rem;
}
 .tailwind .overflow-x-auto {
	 overflow-x: auto;
}
 .tailwind .whitespace-nowrap {
	 white-space: nowrap;
}
 .tailwind .rounded-lg {
	 border-radius: 0.5rem;
}
 .tailwind .rounded-md {
	 border-radius: 0.375rem;
}
 .tailwind .rounded-sm {
	 border-radius: 0.125rem;
}
 .tailwind .rounded-b-lg {
	 border-bottom-right-radius: 0.5rem;
	 border-bottom-left-radius: 0.5rem;
}
 .tailwind .border {
	 border-width: 1px;
}
 .tailwind .border-b {
	 border-bottom-width: 1px;
}
 .tailwind .border-t {
	 border-top-width: 1px;
}
 .tailwind .border-gray-200 {
  --tw-border-opacity: 1;
  border-color: rgb(229 231 235 / var(--tw-border-opacity))
}
 .tailwind .border-slate-200 {
	 --tw-border-opacity: 1;
    border-color: rgb(226 232 240 / var(--tw-border-opacity))
}
 .tailwind .bg-black\/70 {
  background-color: rgb(0 0 0 / 0.7)
}
 .tailwind .bg-slate-200 {
	 --tw-bg-opacity: 1;
  background-color: rgb(226 232 240 / var(--tw-bg-opacity))
}
 .tailwind .bg-slate-50 {
	 --tw-bg-opacity: 1;
  background-color: rgb(248 250 252 / var(--tw-bg-opacity))
}
 .tailwind .bg-white {
	 --tw-bg-opacity: 1;
  background-color: rgb(255 255 255 / var(--tw-bg-opacity))
}
 .tailwind .bg-gradient-to-r {
   background-image: linear-gradient(to right, var(--tw-gradient-stops))
}
 .tailwind .from-transparent {
  --tw-gradient-from: transparent var(--tw-gradient-from-position);
  --tw-gradient-to: rgb(0 0 0 / 0) var(--tw-gradient-to-position);
  --tw-gradient-stops: var(--tw-gradient-from), var(--tw-gradient-to)
}
 .tailwind .to-white {
  --tw-gradient-to: #fff var(--tw-gradient-to-position)
}
 .tailwind .p-3 {
	 padding: 0.75rem;
}
 .tailwind .px-2 {
	 padding-left: 0.5rem;
	 padding-right: 0.5rem;
}
 .tailwind .px-3 {
	 padding-left: 0.75rem;
	 padding-right: 0.75rem;
}
 .tailwind .py-1 {
	 padding-top: 0.25rem;
	 padding-bottom: 0.25rem;
}
 .tailwind .py-1\.5 {
	 padding-top: 0.375rem;
	 padding-bottom: 0.375rem;
}
 .tailwind .py-2 {
	 padding-top: 0.5rem;
	 padding-bottom: 0.5rem;
}
 .tailwind .py-2\.5 {
	 padding-top: 0.625rem;
	 padding-bottom: 0.625rem;
}
 .tailwind .pl-8 {
	 padding-left: 2rem;
}
 .tailwind .pl-\[1\.2rem\] {
	 padding-left: 1.2rem;
}
 .tailwind .text-left {
	 text-align: left;
}
 .tailwind .text-center {
	 text-align: center;
}
 .tailwind .text-\[\.6rem\] {
	 font-size: 0.6rem;
}
 .tailwind .text-\[\.7rem\] {
	 font-size: 0.7rem;
}
 .tailwind .text-\[\.8rem\] {
	 font-size: 0.8rem;
}
 .tailwind .text-\[\.9rem\] {
	 font-size: 0.9rem;
}
 .tailwind .text-\[1\.3rem\] {
	 font-size: 1.3rem;
}
 .tailwind .text-\[1\.5rem\] {
	 font-size: 1.5rem;
}
 .tailwind .text-\[12px\] {
	 font-size: 12px;
}
 .tailwind .text-sm {
	 font-size: 0.875rem;
	 line-height: 1.25rem;
}
 .tailwind .text-xs {
	 font-size: 0.75rem;
	 line-height: 1rem;
}
 .tailwind .font-normal {
	 font-weight: 400;
}
 .tailwind .\!text-slate-200 {
	 --tw-text-opacity: 1 !important;
  color: rgb(226 232 240 / var(--tw-text-opacity)) !important
}
 .tailwind .text-gray-900 {
	 --tw-text-opacity: 1;
  color: rgb(17 24 39 / var(--tw-text-opacity))
}
 .tailwind .text-slate-400 {
	 --tw-text-opacity: 1;
  color: rgb(148 163 184 / var(--tw-text-opacity))
}
 .tailwind .text-slate-500 {
	 --tw-text-opacity: 1;
  color: rgb(100 116 139 / var(--tw-text-opacity))
}
 .tailwind .opacity-0 {
	 opacity: 0;
}
 .tailwind .opacity-100 {
	 opacity: 1;
}
 .tailwind .opacity-50 {
	 opacity: 0.5;
}
 .tailwind .shadow {
  --tw-shadow: 0 1px 3px 0 rgb(0 0 0 / 0.1), 0 1px 2px -1px rgb(0 0 0 / 0.1);
  --tw-shadow-colored: 0 1px 3px 0 var(--tw-shadow-color), 0 1px 2px -1px var(--tw-shadow-color);
  box-shadow: var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)
}
 .tailwind .shadow-\[inset_0_-2px_0_\#dfe6ec\] {
  --tw-shadow: inset 0 -2px 0 #dfe6ec;
  --tw-shadow-colored: inset 0 -2px 0 var(--tw-shadow-color);
  box-shadow: var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)
}
 .tailwind .shadow-sm {
  --tw-shadow: 0 1px 2px 0 rgb(0 0 0 / 0.05);
  --tw-shadow-colored: 0 1px 2px 0 var(--tw-shadow-color);
  box-shadow: var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)
}
 .tailwind .transition-all {
	 transition-property: all;
	 transition-timing-function: cubic-bezier(0.4, 0, 0.2, 1);
	 transition-duration: 150ms;
}
 .tailwind .transition-opacity {
	 transition-property: opacity;
	 transition-timing-function: cubic-bezier(0.4, 0, 0.2, 1);
	 transition-duration: 150ms;
}
 .tailwind .duration-300 {
	 transition-duration: 300ms;
}
 .tailwind .placeholder\:text-gray-400::placeholder {
	 --tw-text-opacity: 1;
  color: rgb(156 163 175 / var(--tw-text-opacity))
}
 .tailwind .first\:sticky:first-child {
	 position: sticky;
}
 .tailwind .first\:left-0:first-child {
	 left: 0px;
}
 .tailwind .first\:z-10:first-child {
	 z-index: 10;
}
 .tailwind .first\:z-20:first-child {
	 z-index: 20;
}
 .tailwind .first\:pl-3:first-child {
	 padding-left: 0.75rem;
}
 .tailwind .first\:text-left:first-child {
	 text-align: left;
}
 .tailwind .first\:shadow-\[inset_-2px_-2px_0_\#dfe6ec\]:first-child {
  --tw-shadow: inset -2px -2px 0 #dfe6ec;
  --tw-shadow-colored: inset -2px -2px 0 var(--tw-shadow-color);
  box-shadow: var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)
}
 .tailwind .first\:shadow-\[inset_-2px_0_0_\#dfe6ec\]:first-child {
  --tw-shadow: inset -2px 0 0 #dfe6ec;
  --tw-shadow-colored: inset -2px 0 0 var(--tw-shadow-color);
  box-shadow: var(--tw-ring-offset-shadow, 0 0 #0000), var(--tw-ring-shadow, 0 0 #0000), var(--tw-shadow)
}
 .tailwind .last\:border-0:last-child {
	 border-width: 0px;
}
 .tailwind .last\:pr-3:last-child {
	 padding-right: 0.75rem;
}
 .tailwind .group:hover .group-hover\:translate-x-2 {
  --tw-translate-x: 0.5rem;
  transform: translate(var(--tw-translate-x), var(--tw-translate-y)) rotate(var(--tw-rotate)) skewX(var(--tw-skew-x)) skewY(var(--tw-skew-y)) scaleX(var(--tw-scale-x)) scaleY(var(--tw-scale-y))
}
 .tailwind .group:hover .group-hover\:bg-slate-50 {
  --tw-bg-opacity: 1;
  background-color: rgb(248 250 252 / var(--tw-bg-opacity))
}
 .tailwind .group:hover .group-hover\:opacity-100 {
	 opacity: 1;
}

.tailwind .text-slate-500 {
  --tw-text-opacity: 1;
  color: rgb(100 116 139/var(--tw-text-opacity));
}

.tailwind .text-secondary-500, .tailwind .text-sky-500 {
    --tw-text-opacity: 1;
    color: rgb(14 165 233/var(--tw-text-opacity));
}
.tailwind .hover\:text-secondary-500:hover {
    --tw-text-opacity: 1;
    color: rgb(14 165 233/var(--tw-text-opacity));
}

.tailwind button i {
  margin-right: 7px;
  margin-left: 7px;
}

.tailwind .bg-yellow-400 {
    --tw-bg-opacity: 1;
    background-color: rgb(250 204 21/var(--tw-bg-opacity));
}

.tailwind .bg-secondary-500 {
    --tw-bg-opacity: 1;
    background-color: rgb(14 165 233/var(--tw-bg-opacity));
}
 @media (min-width: 768px) {
	 .tailwind .md\:ml-0 {
		 margin-left: 0px;
	}
	 .tailwind .md\:hidden {
		 display: none;
	}
	 .tailwind .md\:w-auto {
		 width: auto;
	}
	 .tailwind .md\:max-w-\[18rem\] {
		 max-width: 18rem;
	}
	 .tailwind .md\:flex-row {
		 flex-direction: row;
	}
	 .tailwind .md\:p-12 {
		 padding: 3rem;
	}
	 .tailwind .md\:px-3 {
		 padding-left: 0.75rem;
		 padding-right: 0.75rem;
	}
	 .tailwind .md\:px-5 {
		 padding-left: 1.25rem;
		 padding-right: 1.25rem;
	}
	 .tailwind .md\:py-4 {
		 padding-top: 1rem;
		 padding-bottom: 1rem;
	}
	 .tailwind .md\:text-\[\.9rem\] {
		 font-size: 0.9rem;
	}
	 .tailwind .md\:first\:pl-6:first-child {
		 padding-left: 1.5rem;
	}
	 .tailwind .md\:last\:pr-6:last-child {
		 padding-right: 1.5rem;
	}
}

.ellipsis {
	overflow: hidden;
	text-overflow: ellipsis;
}

.row-radio .el-radio__inner {
	border-width: 1px;
}

.tailwind .el-checkbox__inner {
	--tw-border-opacity: 1;
	border-width: 1px;
	border-color: rgb(226 232 240 / var(--tw-border-opacity));
}
.tailwind .el-checkbox__input.is-checked .el-checkbox__inner,
.tailwind .el-checkbox__input.is-indeterminate .el-checkbox__inner {
	--tw-bg-opacity: 1;
	background-color: rgb(14 165 233 / var(--tw-bg-opacity));
	border-color: rgb(14 165 233 / var(--tw-bg-opacity));
}
/* refine the checkmark so it's thinner and not a full box */
.tailwind .el-checkbox__inner::after {
	border-left: 0 !important;
	border-top: 0 !important;
}
.tailwind .el-checkbox__input.is-checked .el-checkbox__inner::after {
	border-right: 1.5px solid #fff;
	border-bottom: 1.5px solid #fff;
	border-left: 0 !important;
	border-top: 0 !important;
}
.tailwind .el-checkbox__input.is-indeterminate .el-checkbox__inner::before {
	background-color: #fff;
	height: 2px;
}

.tailwind .el-checkbox__inner::after {
	left: 4px;
	top: 1px;
	width: 4px;
	height: 8px;
}

.sly-fast-table-disable-overlay {
	position: absolute;
	top: 0px;
	left: 0px;
	right: 0px;
	bottom: 0px;
	z-index: 2000;
	background-color: rgba(227, 230, 236, 0.5);
	border-radius: 0.5rem;
}