# coding: utf-8

# fields names
ID = 'id'
KEY = 'key'
# INDEX = 'index'
FIGURES = 'figures'
OBJECT_ID = 'objectId'
OBJECT_KEY = 'objectKey'
TAGS = 'tags'

TAGS_MAP = 'tags'
OBJECTS_MAP = 'objects'
FIGURES_MAP = 'figures'
POINTCLOUD_MAP = 'pointclouds'

OBJECTS = 'objects'
POINTCLOUD_ID = 'pointCloudId'
NAME = 'name'
TAGS = 'tags'
DESCRIPTION = 'description'
