# coding: utf-8

# fields names
ALPHA_MASK = "alpha_mask"
BITMAP = "bitmap"
MASK_3D = "mask_3d"
COLS = "cols"
DATA = "data"
EXTERIOR = "exterior"
FACES = "faces"
INTERIOR = "interior"
MULTICHANNEL_BITMAP = "multichannelBitmap"
ORIGIN = "origin"
SPACE_ORIGIN = "space_origin"
SPACE = "space"
SPACE_DIRECTIONS = "space_directions"
POINTS = "points"
ROWS = "rows"
TYPE = "type"
GEOMETRY_SHAPE = "shape"
GEOMETRY_TYPE = "geometryType"
ANY_SHAPE = "any"

NODES = "nodes"
EDGES = "edges"
ENABLED = "enabled"
LABEL = "label"
COLOR = "color"
TEMPLATE = "template"
LOCATION = "location"
LOC = "loc"

X = "x"
Y = "y"
Z = "z"
POSITION = "position"
ROTATION = "rotation"
DIMENTIONS = "dimensions"


LABELER_LOGIN = "labelerLogin"
UPDATED_AT = "updatedAt"
CREATED_AT = "createdAt"
ID = "id"
CLASS_ID = "classId"

INDICES = "indices"

TRACK_ID = "trackId"
